import 'dart:async';
import 'dart:ui';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_cache_manager/flutter_cache_manager.dart';

import 'stream-completer.dart';

class SmoothImageProvider extends ImageProvider<SmoothImageProvider> {
  const SmoothImageProvider(
    this.url, {
    this.scale = 1.0,
    this.headers,
    this.cacheManager,
  })  : assert(url != null),
        assert(scale != null);

  final BaseCacheManager cacheManager;

  final String url;

  final double scale;

  final Map<String, String> headers;

  @override
  ImageStreamCompleter load(SmoothImageProvider key, decode) {
    final StreamController<ImageChunkEvent> chunkEvents =
        StreamController<ImageChunkEvent>();
    return MultiImageStreamCompleter(
      codec: _loadAsync(key, chunkEvents, decode),
      chunkEvents: chunkEvents.stream,
      scale: key.scale,
      informationCollector: () sync* {
        yield DiagnosticsProperty<ImageProvider>(
          'Image provider: $this \n Image key: $key',
          this,
          style: DiagnosticsTreeStyle.errorProperty,
        );
      },
    );
  }

  @override
  Future<SmoothImageProvider> obtainKey(ImageConfiguration configuration) {
    return SynchronousFuture<SmoothImageProvider>(this);
  }

  Stream<Codec> _loadAsync(
    SmoothImageProvider key,
    StreamController<ImageChunkEvent> chunkEvents,
    DecoderCallback decode,
  ) async* {
    assert(key == this);
    try {
      var cm = cacheManager ?? DefaultCacheManager();
      await for (var result
          in cm.getFileStream(key.url, withProgress: true, headers: headers)) {
        if (result is DownloadProgress) {
          chunkEvents.add(ImageChunkEvent(
            cumulativeBytesLoaded: result.downloaded,
            expectedTotalBytes: result.totalSize,
          ));
        }
        if (result is FileInfo) {
          var file = result.file;
          var bytes = await file.readAsBytes();
          var decoded = await decode(bytes);
          yield decoded;
        }
      }
    } catch (e) {
      rethrow;
    } finally {
      await chunkEvents.close();
    }
  }

  @override
  bool operator ==(dynamic other) {
    if (other is SmoothImageProvider) {
      return url == other.url && scale == other.scale;
    }
    return false;
  }

  @override
  int get hashCode => hashValues(url, scale);

  @override
  String toString() => '$runtimeType("$url", scale: $scale)';
}
