import 'package:flutter/cupertino.dart';
import 'package:flutter_cache_manager/flutter_cache_manager.dart';

import 'fade_widget.dart';
import 'image-interface.dart';
import 'image-provoder.dart';

enum _PlaceholderType {
  NONE,
  STATIC,
  PROGRESS,
}

class SmoothImage extends StatefulWidget {
  final ImageProvider image;

  final String imageUrl;

  final SmoothImageErrorWidgetBuilder errorBuilder;

  final ImagePlaceHolderBuilder placeHolderBuilder;

  final ImageProgressIndicatorBuilder progressIndicatorBuilder;

  final ImageTransformer imageBuilder;

  final Duration fadeDuration;

  final double width;

  final double height;

  final Color color;

  final FilterQuality filterQuality;

  final BlendMode colorBlendMode;

  final BoxFit fit;

  final AlignmentGeometry alignment;

  final ImageRepeat repeat;

  final Rect centerSlice;

  final bool matchTextDirection;

  final bool gaplessPlayback;

  final String semanticLabel;

  final bool excludeFromSemantics;

  final bool isAntiAlias;

  SmoothImage({
    Key key,
    @required this.imageUrl,
    Map<String, String> httpHeaders,
    this.imageBuilder,
    this.placeHolderBuilder,
    this.progressIndicatorBuilder,
    this.errorBuilder,
    BaseCacheManager cacheManager,
    this.fadeDuration =const Duration(milliseconds: 500),
    this.semanticLabel,
    this.excludeFromSemantics = false,
    this.width,
    this.height,
    this.color,
    this.colorBlendMode,
    this.fit,
    this.alignment = Alignment.center,
    this.repeat = ImageRepeat.noRepeat,
    this.centerSlice,
    this.matchTextDirection = false,
    this.gaplessPlayback = false,
    this.isAntiAlias = false,
    this.filterQuality = FilterQuality.low,
    int cacheWidth,
    int cacheHeight,
  })  : assert(imageUrl != null),
        assert(alignment != null),
        assert(repeat != null),
        assert(filterQuality != null),
        assert(matchTextDirection != null),
        assert(isAntiAlias != null),
        assert(cacheWidth==null||cacheWidth>0),
        assert(cacheHeight==null||cacheHeight>0),
        this.image = ResizeImage.resizeIfNeeded(
            cacheWidth,
            cacheHeight,
            SmoothImageProvider(
              imageUrl,
              headers: httpHeaders,
              cacheManager: cacheManager,
            )),
        super(key: key);

  @override
  State<StatefulWidget> createState() => _SmoothImageState();
}

class _SmoothImageState extends State<SmoothImage> {
  Widget _previousImage;
  Widget _resolvedImage;
  bool _wasSynchronouslyLoaded = false;
  bool _isLoaded = false;

  @override
  void didUpdateWidget(SmoothImage oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.image != widget.image) {
      if (widget.gaplessPlayback && _resolvedImage != null) {
        _previousImage = _resolvedImage;
      }
      _resolvedImage = null;
    }
  }

  @override
  Widget build(BuildContext context) {
    final type = _definePlaceholderType();
    ImageFrameBuilder frameBuilder;
    switch (type) {
      case _PlaceholderType.NONE:
        frameBuilder = _imageBuilder;
        break;
      case _PlaceholderType.PROGRESS:
        frameBuilder = _preLoadingBuilder;
        break;
      case _PlaceholderType.STATIC:
        frameBuilder = _placeholderBuilder;
        break;
    }
    return Image(
      image: widget.image,
      loadingBuilder:
          type == _PlaceholderType.PROGRESS ? _loadingBuilder : null,
      frameBuilder: frameBuilder,
      errorBuilder: widget.errorBuilder == null ? null : _errorBuilder,
      fit: widget.fit,
      semanticLabel: widget.semanticLabel,
      excludeFromSemantics: widget.excludeFromSemantics,
      width: widget.width,
      height: widget.height,
      alignment: widget.alignment,
      repeat: widget.repeat,
      color: widget.color,
      colorBlendMode: widget.colorBlendMode,
      matchTextDirection: widget.matchTextDirection,
      filterQuality: widget.filterQuality,
      centerSlice: widget.centerSlice,
      gaplessPlayback: widget.gaplessPlayback,
      isAntiAlias: widget.isAntiAlias,
    );
  }

  Widget _errorBuilder(
    BuildContext context,
    Object error,
    StackTrace stackTrace,
  ) {
    return widget.errorBuilder.call(context, widget.imageUrl, error);
  }

  _PlaceholderType _definePlaceholderType() {
    assert(widget.placeHolderBuilder == null ||
        widget.progressIndicatorBuilder == null);
    if (_previousImage != null) return _PlaceholderType.STATIC;
    if (widget.placeHolderBuilder != null) return _PlaceholderType.STATIC;
    if (widget.progressIndicatorBuilder != null) {
      return _PlaceholderType.PROGRESS;
    }
    return _PlaceholderType.NONE;
  }

  Widget _loadingBuilder(
      BuildContext context, Widget child, ImageChunkEvent loadingProgress) {

    if (_isLoaded) {
      final image = _image(context, child);
      if (_wasSynchronouslyLoaded ||
          widget.fadeDuration == null ||
          widget.fadeDuration == Duration.zero) {
        return image;
      }
      final progress =
          widget.progressIndicatorBuilder.call(context,_makeDownloadProgress(loadingProgress));
      return  integrationAnimationWidget(image, progress);
    }
    return widget.progressIndicatorBuilder.call(context, _makeDownloadProgress(loadingProgress));
  }

  Widget integrationAnimationWidget(Widget shown,Widget disappear){
    return Stack(
      fit: StackFit.passthrough,
      alignment: Alignment.center,
      children: [
        FadeWidget(
          child: shown,
          duration: widget.fadeDuration,
        ),
        FadeWidget(
          child: disappear,
          duration: widget.fadeDuration,
          direction: AnimationDirection.reverse,
        )
      ],
    );
  }

  DownloadProgress _makeDownloadProgress(ImageChunkEvent loadingProgress){
    int totalSize;
    int downloaded = 0;
    if (loadingProgress != null) {
      totalSize = loadingProgress.expectedTotalBytes;
      downloaded = loadingProgress.cumulativeBytesLoaded;
    }
    return DownloadProgress(widget.imageUrl,totalSize,downloaded);
  }



  Widget _placeholderBuilder(BuildContext context, Widget child, int frame,
      bool wasSynchronouslyLoaded) {
    final placeholder = _placeholder(context);
    if (frame == null) {
      return placeholder;
    }
    final image = _image(context, child);
    if (wasSynchronouslyLoaded) {
      return image;
    }
    if (widget.fadeDuration == null || widget.fadeDuration == Duration.zero) {
      return image;
    }
    return integrationAnimationWidget(image, placeholder);
  }

  Widget _stackLayoutBuilder(Widget topChild, Key topChildKey,
      Widget bottomChild, Key bottomChildKey) {
    return Stack(
      clipBehavior: Clip.none,
      alignment: Alignment.center,
      fit: StackFit.passthrough,
      children: <Widget>[
        bottomChild,
        topChild,
      ],
    );
  }

  Widget _placeholder(BuildContext context) {
    if (_previousImage != null) return _previousImage;

    if (widget.placeHolderBuilder != null) {
      return widget.placeHolderBuilder(context,widget.imageUrl);
    }
    return SizedBox.shrink();
  }

  Widget _imageBuilder(BuildContext context, Widget child, int frame,
      bool wasSynchronouslyLoaded) {
    if (frame == null) {
      return child;
    }
    return _image(context, child);
  }

  Widget _image(BuildContext context, Widget child) {
    if (widget.imageBuilder != null) {
      _resolvedImage = widget.imageBuilder(context, child,widget.image);
    } else {
      _resolvedImage = child;
    }
    return _resolvedImage;
  }

  Widget _preLoadingBuilder(BuildContext context, Widget child, int frame,
      bool wasSynchronouslyLoaded) {
    _wasSynchronouslyLoaded = wasSynchronouslyLoaded;
    _isLoaded = frame != null;
    return child;
  }
}
