import 'dart:async';
import 'dart:ui';

import 'package:flutter/foundation.dart';
import 'package:flutter/scheduler.dart';
import 'package:flutter/widgets.dart';

class MultiImageStreamCompleter extends ImageStreamCompleter {
  MultiImageStreamCompleter(
      {@required Stream<Codec> codec,
      @required this.scale,
      Stream<ImageChunkEvent> chunkEvents,
      this.informationCollector})
      : assert(codec != null) {
    codec.listen((event) {
      if (_timer != null) {
        _nextImageCodec = event;
      } else {
        _handleCodecReady(event);
      }
    }, onError: (dynamic error, StackTrace stack) {
      reportError(
          context: ErrorDescription('resolving an image codec'),
          exception: error,
          stack: stack,
          informationCollector: informationCollector,
          silent: true);
    });
    if(chunkEvents!=null){
      chunkEvents.listen(reportImageChunkEvent,onError: (dynamic error, StackTrace stack) {
        reportError(
          context: ErrorDescription('loading an image'),
          exception: error,
          stack: stack,
          informationCollector: informationCollector,
          silent: true,
        );
      });
    }
  }

  final double scale;
  final InformationCollector informationCollector;
  Codec _codec;
  Codec _nextImageCodec;
  FrameInfo _nextFrame;
  Duration _shownTimestamp;
  Duration _frameDuration;
  int _framesEmitted = 0;
  Timer _timer;
  bool _frameCallbackScheduled = false;

  void _handleCodecReady(Codec codec) {
    _codec = codec;
    assert(_codec != null);

    if (hasListeners) {
      _decodeNextFrameAndSchedule();
    }
  }

  Future<void> _decodeNextFrameAndSchedule() async {
    try {
      _nextFrame = await _codec.getNextFrame();
    } catch (exception, stack) {
      reportError(
        context: ErrorDescription('resolving an image frame'),
        exception: exception,
        stack: stack,
        informationCollector: informationCollector,
        silent: true,
      );
      return;
    }
    if (_codec.frameCount == 1) {
      // This is not an animated image, just return it and don't schedule more
      // frames.
      _emitFrame(ImageInfo(image: _nextFrame.image, scale: scale));
      return;
    }
    _scheduleAppFrame();
  }

  void _emitFrame(ImageInfo imageInfo) {
    setImage(imageInfo);
    _framesEmitted += 1;
  }

  void _scheduleAppFrame() {
    if (_frameCallbackScheduled) {
      return;
    }
    _frameCallbackScheduled = true;
    SchedulerBinding.instance.scheduleFrameCallback(_handleAppFrame);
  }

  void _handleAppFrame(Duration timestamp) {
    _frameCallbackScheduled = false;
    if (!hasListeners) return;
    if (_isFirstFrame() || _hasFrameDurationPassed(timestamp)) {
      _emitFrame(ImageInfo(image: _nextFrame.image, scale: scale));
      _shownTimestamp = timestamp;
      _frameDuration = _nextFrame.duration;
      _nextFrame = null;
      if (_framesEmitted % _codec.frameCount == 0 && _nextImageCodec != null) {
        _switchToNewCodec();
      } else {
        final int completedCycles = _framesEmitted ~/ _codec.frameCount;
        if (_codec.repetitionCount == -1 ||
            completedCycles <= _codec.repetitionCount) {
          _decodeNextFrameAndSchedule();
        }
      }
      return;
    }
    final Duration delay = _frameDuration - (timestamp - _shownTimestamp);
    _timer = Timer(delay * timeDilation, _scheduleAppFrame);
  }

  bool _isFirstFrame() {
    return _frameDuration == null;
  }

  bool _hasFrameDurationPassed(Duration timestamp) {
    assert(_shownTimestamp != null);
    return timestamp - _shownTimestamp >= _frameDuration;
  }

  void _switchToNewCodec() {
    _framesEmitted = 0;
    _timer = null;
    _handleCodecReady(_nextImageCodec);
    _nextImageCodec = null;
  }

  @override
  void addListener(ImageStreamListener listener) {
    if(!hasListeners&&_codec!=null){
      _decodeNextFrameAndSchedule();
    }
    super.addListener(listener);
  }

  @override
  void removeListener(ImageStreamListener listener) {
    super.removeListener(listener);
    if (!hasListeners) {
      _timer?.cancel();
      _timer = null;
    }
  }

}
